"""
Support for request forwarding.

HATEOAS depends heavily on services being able to generate URIs back to their own resources.
Under some request forwarding scenarios (especially AWS ALBs), services may not be resolved
on a URI that is accessible to other services. This can be solved by configuring static
service URIs... or by resolving URIs using X-Forwarded headers.

"""
from flask import request, _request_ctx_stack

from microcosm_flask.session import register_session_factory


def use_forwarded_port(graph):
    """
    Inject the `X-Forwarded-Port` (if any) into the current URL adapter.

    The URL adapter is used by `url_for` to build a URLs.

    """
    forwarded_port = request.headers.get("X-Forwarded-Port")
    if not forwarded_port:
        return None

    if _request_ctx_stack is None:
        return None

    # There must be a better way!
    context = _request_ctx_stack.top
    context.url_adapter.server_name = "{}:{}".format(
        context.url_adapter.server_name.split(":", 1)[0],
        forwarded_port,
    )

    return forwarded_port


def configure_port_forwarding(graph):
    """
    Bind the SQLAlchemy session context to Flask.

    The current session is available at `g.db.session`.

    """
    return register_session_factory(graph, "forwarded_port", use_forwarded_port)



"""
Conventions for canonical CRUD endpoints.

"""
from inflection import pluralize
from marshmallow import Schema

from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import (
    dump_response_data,
    encode_count_header,
    load_query_string_data,
    load_request_data,
    merge_data,
    require_response_data,
)
from microcosm_flask.conventions.registry import qs, request, response
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation
from microcosm_flask.paging import Page, PaginatedList, make_paginated_list_schema


class CRUDConvention(Convention):

    @property
    def page_cls(self):
        return Page

    def configure_search(self, ns, definition):
        """
        Register a search endpoint.

        The definition's func should be a search function, which must:
        - accept kwargs for the query string (minimally for pagination)
        - return a tuple of (items, count) where count is the total number of items
          available (in the case of pagination)

        The definition's request_schema will be used to process query string arguments.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        paginated_list_schema = make_paginated_list_schema(ns, definition.response_schema)()

        @self.graph.route(ns.collection_path, Operation.Search, ns)
        @qs(definition.request_schema)
        @response(paginated_list_schema)
        def search(**path_data):
            request_data = load_query_string_data(definition.request_schema)
            page = self.page_cls.from_query_string(request_data)
            return_value = definition.func(**merge_data(path_data, page.to_dict(as_str=False)))

            if len(return_value) == 3:
                items, count, context = return_value
            else:
                context = {}
                items, count = return_value

            response_data = PaginatedList(
                ns=ns,
                page=page,
                items=items,
                count=count,
                schema=definition.response_schema,
                operation=Operation.Search,
                **context
            )

            headers = encode_count_header(count)
            return dump_response_data(paginated_list_schema, response_data, headers=headers)

        search.__doc__ = "Search the collection of all {}".format(pluralize(ns.subject_name))

    def configure_count(self, ns, definition):
        """
        Register a count endpoint.

        The definition's func should be a count function, which must:
        - accept kwargs for the query string
        - return a count is the total number of items available

        The definition's request_schema will be used to process query string arguments.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.graph.route(ns.collection_path, Operation.Count, ns)
        @qs(definition.request_schema)
        def count(**path_data):
            request_data = load_query_string_data(definition.request_schema)
            count = definition.func(**merge_data(path_data, request_data))
            headers = encode_count_header(count)
            return dump_response_data(None, None, headers=headers)

        count.__doc__ = "Count the size of the collection of all {}".format(pluralize(ns.subject_name))

    def configure_create(self, ns, definition):
        """
        Register a create endpoint.

        The definition's func should be a create function, which must:
        - accept kwargs for the request and path data
        - return a new item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.graph.route(ns.collection_path, Operation.Create, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def create(**path_data):
            request_data = load_request_data(definition.request_schema)
            response_data = definition.func(**merge_data(path_data, request_data))
            return dump_response_data(definition.response_schema, response_data, Operation.Create.value.default_code)

        create.__doc__ = "Create a new {}".format(ns.subject_name)

    def configure_updatebatch(self, ns, definition):
        """
        Register an update batch endpoint.

        The definition's func should be an update function, which must:
        - accept kwargs for the request and path data
        - return a new item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        operation = Operation.UpdateBatch

        @self.graph.route(ns.collection_path, operation, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def update_batch(**path_data):
            request_data = load_request_data(definition.request_schema)
            response_data = definition.func(**merge_data(path_data, request_data))
            return dump_response_data(definition.response_schema, response_data, operation.value.default_code)

        update_batch.__doc__ = "Update a batch of {}".format(ns.subject_name)

    def configure_retrieve(self, ns, definition):
        """
        Register a retrieve endpoint.

        The definition's func should be a retrieve function, which must:
        - accept kwargs for path data
        - return an item or falsey

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        request_schema = definition.request_schema or Schema()

        @self.graph.route(ns.instance_path, Operation.Retrieve, ns)
        @qs(request_schema)
        @response(definition.response_schema)
        def retrieve(**path_data):
            request_data = load_query_string_data(request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(definition.response_schema, response_data)

        retrieve.__doc__ = "Retrieve a {} by id".format(ns.subject_name)

    def configure_delete(self, ns, definition):
        """
        Register a delete endpoint.

        The definition's func should be a delete function, which must:
        - accept kwargs for path data
        - return truthy/falsey

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.graph.route(ns.instance_path, Operation.Delete, ns)
        def delete(**path_data):
            require_response_data(definition.func(**path_data))
            return "", Operation.Delete.value.default_code

        delete.__doc__ = "Delete a {} by id".format(ns.subject_name)

    def configure_replace(self, ns, definition):
        """
        Register a replace endpoint.

        The definition's func should be a replace function, which must:
        - accept kwargs for the request and path data
        - return the replaced item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.graph.route(ns.instance_path, Operation.Replace, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def replace(**path_data):
            request_data = load_request_data(definition.request_schema)
            # Replace/put should create a resource if not already present, but we do not
            # enforce these semantics at the HTTP layer. If `func` returns falsey, we
            # will raise a 404.
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(definition.response_schema, response_data)

        replace.__doc__ = "Create or update a {} by id".format(ns.subject_name)

    def configure_update(self, ns, definition):
        """
        Register an update endpoint.

        The definition's func should be an update function, which must:
        - accept kwargs for the request and path data
        - return an updated item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.graph.route(ns.instance_path, Operation.Update, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def update(**path_data):
            # NB: using partial here means that marshmallow will not validate required fields
            request_data = load_request_data(definition.request_schema, partial=True)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(definition.response_schema, response_data)

        update.__doc__ = "Update some or all of a {} by id".format(ns.subject_name)


def configure_crud(graph, ns, mappings, path_prefix=""):
    """
    Register CRUD endpoints for a resource object.

    :param mappings: a dictionary from operations to tuple, where each tuple contains
                     the target function and zero or more marshmallow schemas according
                     to the signature of the "register_<foo>_endpoint" functions

    Example mapping:

        {
            Operation.Create: (create_foo, NewFooSchema(), FooSchema()),
            Operation.Delete: (delete_foo,),
            Operation.Retrieve: (retrieve_foo, FooSchema()),
        }

    """
    ns = Namespace.make(ns, path=path_prefix)
    convention = CRUDConvention(graph)
    convention.configure(ns, mappings)



"""
Pagination support.

"""
from flask import request
from marshmallow import fields, Schema

from microcosm_flask.linking import Link, Links
from microcosm_flask.operations import Operation


def identity(x):
    """
    Identity function.

    """
    return x


class PageSchema(Schema):
    offset = fields.Integer(missing=None)
    limit = fields.Integer(missing=None)


def make_paginated_list_schema(ns, item_schema):
    """
    Generate a paginated list schema.

    :param ns: a `Namespace` for the list's item type
    :param item_schema: a `Schema` for the list's item type

    """

    class PaginatedListSchema(Schema):
        __alias__ = "{}_list".format(ns.subject_name)

        offset = fields.Integer(required=True)
        limit = fields.Integer(required=True)
        count = fields.Integer(required=True)
        items = fields.List(fields.Nested(item_schema), required=True)
        _links = fields.Raw()

    return PaginatedListSchema


class Page(object):

    def __init__(self, offset=None, limit=None, **rest):
        self.offset = self.default_offset if offset is None else offset
        self.limit = self.default_limit if limit is None else limit
        self.rest = rest

    @property
    def default_offset(self):
        return 0

    @property
    def default_limit(self):
        try:
            return int(request.headers["X-Request-Limit"])
        except:
            return 20

    @classmethod
    def from_query_string(cls, qs):
        """
        Create a page from a query string dictionary.

        This dictionary should probably come from `PageSchema.from_request()`.

        """
        dct = qs.copy()
        offset = dct.pop("offset", None)
        limit = dct.pop("limit", None)
        return cls(
            offset=offset,
            limit=limit,
            **dct
        )

    def next(self):
        return Page(
            offset=self.offset + self.limit,
            limit=self.limit,
            **self.rest
        )

    def prev(self):
        return Page(
            offset=self.offset - self.limit,
            limit=self.limit,
            **self.rest
        )

    def to_dict(self, as_str=False):
        return dict(self.to_tuples(as_str=as_str))

    def to_tuples(self, as_str=True):
        """
        Convert to tuples for deterministic order when passed to urlencode.

        """
        value_func = str if as_str else identity

        return [
            ("offset", self.offset),
            ("limit", self.limit),
        ] + [
            (key, value_func(self.rest[key]))
            for key in sorted(self.rest.keys())
        ]


class PaginatedList(object):

    def __init__(self,
                 ns,
                 page,
                 items,
                 count,
                 schema=None,
                 operation=Operation.Search,
                 **extra):
        self.ns = ns
        self.page = page
        self.items = items
        self.count = count
        self.schema = schema
        self.operation = operation
        self.extra = extra

    def to_dict(self):
        return dict(
            count=self.count,
            items=[
                self.schema.dump(item).data if self.schema else item
                for item in self.items
            ],
            _links=self._links,
            **self.page.to_dict(as_str=True)
        )

    @property
    def offset(self):
        return self.page.offset

    @property
    def limit(self):
        return self.page.limit

    @property
    def _links(self):
        return self.links.to_dict()

    @property
    def links(self):
        links = Links()
        links["self"] = Link.for_(self.operation, self.ns, qs=self.page.to_tuples(), **self.extra)
        if self.page.offset + self.page.limit < self.count:
            links["next"] = Link.for_(self.operation, self.ns, qs=self.page.next().to_tuples(), **self.extra)
        if self.page.offset > 0:
            links["prev"] = Link.for_(self.operation, self.ns, qs=self.page.prev().to_tuples(), **self.extra)
        return links



"""
Support for encoding and decoding request/response content.

"""
from flask import jsonify, request
from werkzeug import Headers
from werkzeug.exceptions import NotFound, UnprocessableEntity


def with_headers(error, headers):
    setattr(error, "headers", headers)
    return error


def with_context(error, errors):
    setattr(error, "context", dict(errors=errors))
    return error


def encode_count_header(count):
    return {
        "X-Total-Count": count,
    }


def load_request_data(request_schema, partial=False):
    """
    Load request data as JSON using the given schema.

    Forces JSON decoding even if the client not specify the `Content-Type` header properly.

    This is friendlier to client and test software, even at the cost of not distinguishing
    HTTP 400 and 415 errors.

    """
    try:
        json_data = request.get_json(force=True) or {}
    except:
        # if `simplpejson` is installed, simplejson.scanner.JSONDecodeError will be raised
        # on malformed JSON, where as built-in `json` returns None
        json_data = {}
    request_data = request_schema.load(json_data, partial=partial)
    if request_data.errors:
        # pass the validation errors back in the context
        raise with_context(
            UnprocessableEntity("Validation error"), [{
                "message": "Could not validate field: {}".format(field),
                "field": field,
                "reasons": reasons
            } for field, reasons in request_data.errors.items()],
        )
    return request_data.data


def load_query_string_data(request_schema):
    """
    Load query string data using the given schema.

    Schemas are assumed to be compatible with the `PageSchema`.

    """
    query_string_data = request.args
    request_data = request_schema.load(query_string_data)
    if request_data.errors:
        # pass the validation errors back in the context
        raise with_context(UnprocessableEntity("Validation error"), dict(errors=request_data.errors))
    return request_data.data


def remove_null_values(data):
    if isinstance(data, dict):
        return {
            key: remove_null_values(value)
            for key, value in data.items()
            if value is not None
        }
    if type(data) in (list, tuple):
        return type(data)(map(remove_null_values, data))
    return data


def dump_response_data(response_schema, response_data, status_code=200, headers=None):
    """
    Dumps response data as JSON using the given schema.

    Forces JSON encoding even if the client did not specify the `Accept` header properly.

    This is friendlier to client and test software, even at the cost of not distinguishing
    HTTP 400 and 406 errors.

    """
    if response_schema:
        response_data = response_schema.dump(response_data).data

    return make_response(response_data, status_code, headers)


def make_response(response_data, status_code=200, headers=None):
    if request.headers.get("X-Response-Skip-Null"):
        # swagger does not currently support null values; remove these conditionally
        response_data = remove_null_values(response_data)

    headers = headers or {}
    if "Content-Type" not in headers:
        # Specify JSON as the response content type by default
        headers["Content-Type"] = "application/json"

    response = jsonify(response_data)
    response.headers = Headers(headers)
    response.status_code = status_code
    return response


def merge_data(path_data, request_data):
    """
    Merge data from the URI path and the request.

    Path data wins.

    """
    merged = request_data.copy() if request_data else {}
    merged.update(path_data or {})
    return merged


def require_response_data(response_data):
    """
    Enforce that response data is truthy.

    Used to automating 404 errors for CRUD functions that return falsey. Does not
    preclude CRUD functions from raising their own errors.

    :raises NotFound: otherwise

    """
    if not response_data:
        raise NotFound
    return response_data



"""
Audit log support for Flask routes.

"""
from collections import namedtuple
from functools import wraps
from logging import getLogger
from json import loads
from traceback import format_exc

from flask import current_app, g, request
from microcosm.api import defaults
from microcosm_flask.errors import (
    extract_context,
    extract_error_message,
    extract_include_stack_trace,
    extract_status_code,
)
from microcosm_logging.timing import elapsed_time


AuditOptions = namedtuple("AuditOptions", [
    "include_request_body",
    "include_response_body",
])


SKIP_LOGGING = "_microcosm_flask_skip_audit_logging"


def skip_logging(func):
    """
    Decorate a function so logging will be skipped.

    """
    setattr(func, SKIP_LOGGING, True)
    return func


def should_skip_logging(func):
    """
    Should we skip logging for this handler?

    """
    return getattr(func, SKIP_LOGGING, False)


def audit(func):
    """
    Record a Flask route function in the audit log.

    Generates a JSON record in the Flask log for every request.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        options = AuditOptions(
            include_request_body=True,
            include_response_body=True,
        )
        return _audit_request(options, func, None, *args, **kwargs)

    return wrapper


class RequestInfo(object):
    """
    Capture of key information for requests.

    """
    def __init__(self, options, func, request_context):
        self.options = options
        self.operation = request.endpoint
        self.func = func.__name__
        self.method = request.method
        self.request_context = request_context
        self.timing = dict()

        self.error = None
        self.stack_trace = None
        self.request_body = None
        self.response_body = None
        self.status_code = None
        self.success = None

    def to_dict(self):
        dct = dict(
            operation=self.operation,
            func=self.func,
            method=self.method,
            **self.timing
        )

        if self.request_context is not None:
            dct.update(self.request_context())

        if self.success is True:
            dct.update(
                success=self.success,
                status_code=self.status_code,
            )
        if self.success is False:
            dct.update(
                success=self.success,
                message=extract_error_message(self.error)[:2048],
                context=extract_context(self.error),
                stack_trace=self.stack_trace,
                status_code=self.status_code,
            )

        self.post_process_request_body(dct)
        self.post_process_response_body(dct)

        return dct

    def log(self, logger):
        if self.status_code == 500:
            # something actually went wrong; investigate
            logger.warning(self.to_dict())
        else:
            # usually log at INFO; a raised exception can be an error or expected behavior (e.g. 404)
            logger.info(self.to_dict())

    def capture_request(self):
        if not current_app.debug:
            # only capture request body on debug
            return

        if not self.options.include_request_body:
            # only capture request body if requested
            return

        if not request.get_json(force=True, silent=True):
            # only capture request body if json
            return

        self.request_body = request.get_json(force=True)

    def capture_response(self, response):
        self.success = True

        body, self.status_code = parse_response(response)

        if not current_app.debug:
            # only capture responsebody on debug
            return

        if not self.options.include_response_body:
            # only capture response body if requested
            return

        if not body:
            # only capture request body if there is one
            return

        try:
            self.response_body = loads(body)
        except (TypeError, ValueError):
            # not json
            pass

    def capture_error(self, error):
        self.error = error
        self.status_code = extract_status_code(error)
        self.success = 0 < self.status_code < 400
        include_stack_trace = extract_include_stack_trace(error)
        self.stack_trace = format_exc(limit=10) if (not self.success and include_stack_trace) else None

    def post_process_request_body(self, dct):
        if g.get("hide_body") or not self.request_body:
            return

        for name, new_name in g.get("show_request_fields", {}).items():
            try:
                value = self.request_body.pop(name)
                self.request_body[new_name] = value
            except KeyError:
                pass

        for field in g.get("hide_request_fields", []):
            try:
                del self.request_body[field]
            except KeyError:
                pass

        dct.update(
            request_body=self.request_body,
        )

    def post_process_response_body(self, dct):
        if g.get("hide_body") or not self.response_body:
            return

        for name, new_name in g.get("show_response_fields", {}).items():
            try:
                value = self.response_body.pop(name)
                self.response_body[new_name] = value
            except KeyError:
                pass

        for field in g.get("hide_response_fields", []):
            try:
                del self.response_body[field]
            except KeyError:
                pass

        dct.update(
            response_body=self.response_body,
        )


def _audit_request(options, func, request_context, *args, **kwargs):  # noqa: C901
    """
    Run a request function under audit.

    """
    logger = getLogger("audit")

    request_info = RequestInfo(options, func, request_context)
    response = None

    request_info.capture_request()
    try:
        # process the request
        with elapsed_time(request_info.timing):
            response = func(*args, **kwargs)
    except Exception as error:
        request_info.capture_error(error)
        raise
    else:
        request_info.capture_response(response)
        return response
    finally:
        if not should_skip_logging(func):
            request_info.log(logger)


def parse_response(response):
    """
    Parse a Flask response into a body and status code.

    The returned value from a Flask view could be:
        * a tuple of (response, status) or (response, status, headers)
        * a Response object
        * a string
    """
    if isinstance(response, tuple) and len(response) > 1:
        return response[0], response[1]
    try:
        return response.data, response.status_code
    except AttributeError:
        return response, 200


@defaults(
    include_request_body=True,
    include_response_body=True,
)
def configure_audit_decorator(graph):
    """
    Configure the audit decorator.

    Example Usage:

        @graph.audit
        def login(username, password):
            ...
    """
    include_request_body = graph.config.audit.include_request_body
    include_response_body = graph.config.audit.include_response_body

    def _audit(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            options = AuditOptions(
                include_request_body=include_request_body,
                include_response_body=include_response_body,
            )
            return _audit_request(options, func, graph.request_context,  *args, **kwargs)
        return wrapper
    return _audit



"""
Support a user-defined per-request session.

"""
from flask import g


def register_session_factory(graph, key, session_factory):
    """
    Register a session creation function so that a new session (of user-defined type)
    will be saved to `flask.g` on every request (and closed on teardown).

    In other words: this os a mechanism to register a SQLAlchemy session instance
    or similar without coupling the web and database tiers directly.

    The session function should have the signature:

        def session_factory(graph):
            return Session()

    If the session instance is closeable, it will be closed on teardown.

    """
    @graph.flask.before_request
    def begin_session():
        setattr(g, key, session_factory(graph))

    @graph.flask.teardown_request
    def end_session(*args, **kwargs):
        # NB: session will be none if there's an error raised in `before_request`
        session = getattr(g, key, None)
        if session is not None and hasattr(session, "close"):
            session.close()



#!/usr/bin/env python
from setuptools import find_packages, setup

project = "microcosm-flask"
version = "0.65.0"

setup(
    name=project,
    version=version,
    description="Opinionated persistence with FlaskQL",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/microcosm-flask",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    keywords="microcosm",
    install_requires=[
        "enum34>=1.1.2",
        "Flask>=0.11",
        "Flask-BasicAuth>=0.2.0",
        "flask-cors>=2.1.2",
        "Flask-UUID>=0.2",
        "marshmallow>=2.12.2",
        "microcosm>=0.12.0",
        "microcosm-logging>=0.12.0",
        "openapi>=0.5.0",
        "python-dateutil>=2.5.2",
        "PyYAML>=3.11",
        "rfc3986>=0.4.1",
    ],
    setup_requires=[
        "nose>=1.3.6",
    ],
    dependency_links=[
    ],
    entry_points={
        "console_scripts": [
            "compare-resources = microcosm_flask.sync.compare:main",
            "sync-resources = microcosm_flask.sync.main:main",
        ],
        "microcosm.factories": [
            "app = microcosm_flask.factories:configure_flask_app",
            "audit = microcosm_flask.audit:configure_audit_decorator",
            "basic_auth = microcosm_flask.basic_auth:configure_basic_auth_decorator",
            "build_info_convention = microcosm_flask.conventions.build_info:configure_build_info",
            "discovery_convention = microcosm_flask.conventions.discovery:configure_discovery",
            "error_handlers = microcosm_flask.errors:configure_error_handlers",
            "flask = microcosm_flask.factories:configure_flask",
            "health_convention = microcosm_flask.conventions.health:configure_health",
            "port_forwarding = microcosm_flask.forwarding:configure_port_forwarding",
            "request_context = microcosm_flask.context:configure_request_context",
            "route = microcosm_flask.routing:configure_route_decorator",
            "swagger_convention = microcosm_flask.conventions.swagger:configure_swagger",
            "uuid = microcosm_flask.converters:configure_uuid",
        ],
    },
    tests_require=[
        "coverage>=3.7.1",
        "mock>=1.0.1",
        "PyHamcrest>=1.8.5",
    ],
)
